/*****************************************************************************
 *
 * FILE:	gps-sub.h
 * DESCRIPTION:	GPS (Global Positioning System) library sub-header file
 * DATE:	Thu, Dec 20 2001
 * UPDATED:	Sat, Sep 20 2003
 * AUTHOR:	Kouichi ABE (WALL)
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2003 Kouichi ABE (WALL), All rights reserved.
 * NOTE:	Supported GPS receiver
 *		1. "SONY PACY-CNV10 (PC-Card)" GPS receiver
 *		2. "DATATEC SPEED SENSOR (GPex)" GPS receiver
 *		   GP/ex SPECIFICATION:
 *			communication	asynchronized full-duplex
 *			speed		4800bps
 *			data		8bit
 *			stop bit	1bit
 *			parity		none
 *			output		ASCII code
 * LICENSE:
 *
 *  Copyright (C) 2001-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: gps-sub.h,v 1.4 2003/09/21 11:16:50 kouichi Exp $
 *
 *****************************************************************************/

#ifndef	_GPS_SUB_H
#define	_GPS_SUB_H	1

#include <stdio.h>
#if	TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif	/* TM_IN_SYS_TIME */


/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#if	HAVE_STRTOL
#define	atoi(x)	(int)strtol((x), (char **)NULL, 10)
#endif	/* HAVE_STRTOL */
#if	HAVE_STRTOD
#define	atof(x)	strtod((x), (char **)NULL)
#endif	/* HAVE_STRTOD */

#define	CR	'\015'
#define	LF	'\012'
#define	CRLF	"\015\012"
#define	EOL	'\0'
#define	SPC	'\040'
#define	TAB	'\011'

#define	RAWDATA_SIZE	(128)
#define	MAX_GSV_ARGS	(8)

#define	GBUFSIZE	(256)
#define	GPS_DEVICE	"/dev/cuaa0"
#define	PACY_OPENSPEED	(B9600)
#define	NMEA_OPENSPEED	(B4800)

#define	DEF_OFFSET	((long)0)	/* 32400:JST */
#define	DEF_TIMEZONE	"GMT"


/*****************************************************************************
 *
 *	NMEA-0183
 *
 *****************************************************************************/
/* GGA - Global Positioning System Fix Data */
struct _gpgga {				/* GPGGA ǡե */
  char		rawdata[RAWDATA_SIZE];	/* NMEA ǡ */
  char		time[6+1];		/* UTC ʻʬá		*/
  char		latitude[11+1];		/* ١1/1,000,000ʬޤǡ	*/
  char		ns;			/* ̰ޡNˡޡS		*/
  char		longitude[12+1];	/* ١1/1,000,000ʬˤޤ	*/
  char		ew;			/* СWˡСE		*/
  char		quality;		/* ʼ 0:Բ 1:ñ¬	*/
					/* 2: ǥե󥷥¬	*/
  int		num;			/* 0008	*/
  double	dop;			/* DOP: PDOP Τ		*/
  double	altitude;		/* ȴ١-999.9m17999.9m*/
  char		altitude_unit;		/* ȴ٤ñ̡'M'	*/
  double	geoid;			/* ɹ⤵m		*/
  char		geoid_unit;		/* ɹ⤵ñ̡'M'*/
  double	dgps_data_age;		/* DGPS ǡ֡á	*/
  char		dgps_station_id[5+1];	/* DGPS ɼֹ		*/
  char		checksum[3+1];		/* å			*/
};

/* GLL - Geographic Position, Latitude and Longitude */
struct _gpgll {				/* GPGLL ǡե */
  char		rawdata[RAWDATA_SIZE];	/* NMEA ǡ */
  char		latitude[11+1];		/* ١1/1,000,000ʬޤǡ	*/
  char		ns;			/* ̰ޡNˡޡS		*/
  char		longitude[12+1];	/* ١1/1,000,000ʬˤޤ	*/
  char		ew;			/* СWˡСE		*/
  char		time[6+3+1];		/* UTC ʻʬ.xx	*/
  char		status;			/* : A = ͭ, V = ̵	*/
  char		mode;	/* ⡼: A = ñ¬, D = DGPS, N = ̵ */
  char		checksum[3+1];		/* å			*/
};

/* GSA - GNSS DOP and Active Satellites */
#define	MAX_GSA_SATELLITES	(11)
struct _gpgsa {				/* GPGSA ǡե */
  char		rawdata[RAWDATA_SIZE];	/* NMEA ǡ */
  char		measure_mode;		/* ¬̥⡼: */
				  /* A = 2D/3Dư, M = 2D/3Dư */
  char		mode;	/* ⡼: 1 = ǽ, 2 = 2D, 3 = 3D */
  int		satellite_no[MAX_GSA_SATELLITES];/* ¬̤Ѥֹ */
  double	pdop;			/* PDOP */
  double	hdop;			/* HDOP */
  double	vdop;			/* VDOP */
  char		checksum[3+1];		/* å			*/
};

/* GSV - Satellites in View */
#define	MAX_GSV_SATELLITES	(4)
struct _gpgsv {				/* GPGSV ǡե */
  char		rawdata[RAWDATA_SIZE];	/* NMEA ǡ */
  int		mesgs;			/* å */
  int		msg_no;			/* åֹ */
  int		satellite_num;		/* ǽ */
  struct {
    int		no;			/* ֹ */
    int		angle;			/* ĳѡ١ */
    int		direction;		/* ѡ١ */
    int		snr;			/* SNRʥǥ٥*/
  } satellites[MAX_GSV_SATELLITES];
  char		checksum[3+1];		/* å			*/
};

/* RMC - Recommended Minimum Specific GNSS Data */
struct _gprmc {				/* GPRMC ǡե */
  char		rawdata[RAWDATA_SIZE];	/* NMEA ǡ */
  char		time[6+1];		/* ¬̻UTCʻʬá	*/
  char		status;			/* : A = ͭ, V = ̵	*/
  char		latitude[11+1];		/* ١1/10,000ʬޤǡ	*/
  char		ns;			/* ̰ޡNˡޡS		*/
  char		longitude[12+1];	/* ١1/10,000ʬˤޤ	*/
  char		ew;			/* СWˡСE		*/
  double	velocity;		/* ®١000.0999.9Υåȡ	*/
  double	direction;		/* ̡000.0359.9 ١	*/
  char		date[6+1];		/* UTC ա++ǯβ	*/
  char		magnetic_deviation[5+1];/* к */
  char		md_revise;		/* к */
  char		mode;		/* ⡼: A = ñ¬, D = DGPS, N = ̵ */
  char		checksum[3+1];		/* å	 */
};

/* VTG - Course Over Ground and Ground Speed */
struct _gpvtg {				/* GPVTG ǡե */
  char		rawdata[RAWDATA_SIZE];	/* NMEA ǡ */
  double	direction;		/* ̡000.0359.9 ١*/
  char		direction_unit;		/* ñ */
  double	magnetic_direction;	/*  */
  char		magnetic_dir_unit;	/* ñ	*/
  double	velocity1;		/* ®١000.0999.9Υåȡ*/
  char		velocity1_unit;		/* ®ñ: Υå */
  double	velocity2;		/* ®١0000.01851.9 km/h*/
  char		velocity2_unit;		/* ®ñ: km/h */
  char		mode;		/* ⡼: A = ñ¬, D = DGPS, N = ̵ */
  char		checksum[3+1];		/* å	 */
};

/* ZDA - Time and Date */
struct _gpzda {			/* GPZDA ǡե */
  char	rawdata[RAWDATA_SIZE];	/* NMEA ǡ */
  char	time[6+1];		/* UTC ʻʬá		*/
  char	day[2+1];		/* UTC : 0131	*/
  char	month[2+1];		/* UTC : 0112	*/
  char	year[4+1];		/* UTC : ǯ19802040	*/
  char	hour[3+1];		/* ϻ֡ʻˡ-13+00+13	*/
				/* ФǤ顢ФǤ	*/
  char	min[2+1];		/* ϻ֡ʬˡ0059	*/
  char	checksum[3+1];		/* å	 */
};

typedef struct {		/* Υǡ¤	*/
  GPSDataFormat	type;
  union {
    struct _pacy {
      char	version[7];	/* ΥС	*/
      struct tm	nowtm;		/* ߻	*/
      char	ns;		/* ̰ޡNˡޡS	*/
      char	ew;		/* СWˡСE	*/
      int	latitude_deg;	/* ١١		*/
      int	latitude_min;	/* ١ʬ: 1/60 deg	*/
      int	latitude_sec;	/* ١á: 1/60 min	*/
      int	latitude_frac;	/* ١ü		*/
      double	n_latitude;	/* ١ʿͲ	*/
      int	longitude_deg;	/* ١١		*/
      int	longitude_min;	/* ١ʬ: 1/60 deg	*/
      int	longitude_sec;	/* ١á: 1/60 min	*/
      int	longitude_frac;	/* ١ü		*/
      double	n_longitude;	/* ١ʿͲ	*/
      int	altitude;	/* ɸ			*/
      int	n_altitude;	/* ɸʿͲ	*/
      int	velocity;	/* ®			*/
      int	direction;	/* 			*/
      struct tm	calctm;		/* ¬	*/
      char *	dop;		/* ١dilution of precision*/
      int	n_dop;		/* ٤ɤ餮	*/
      int	measuremode;	/* ¬⡼ɡ2-D or 3-D	*/
      char *	mapdatum;	/* ǡ	*/
      struct _channel {		/* ͥ빽¤	*/
	int	number;		/* ֹͥ		*/
	char	status[6];	/* ͥξ	*/
	int	prn;		/* PRN ֹ桩		*/
	int	n_elevation;	/* ι١ʿ͡	*/
	char	elevation[12];	/* : "-(xxx..xxx)"	*/
	int	n_azimuth;	/* ̡ʿ͡	*/
	char	azimuth[9];	/* : "-(xx..xx)"	*/
	char *	situation;	/* ȱξ	*/
	int	n_situation;	/* ȱξ	*/
	char	c_situation;	/* ȱξ	*/
	int	signal;		/* ξ		*/
      } ch[8];			/* Υͥ8ġ*/
      char	rest[5];	/* ʥǡ		*/
      char	parity;		/* ѥƥeither E or 0	*/
    } pacy;
    struct _nmea {
      struct _gpgga	gga;
      struct _gpgll	gll;
      struct _gpgsa	gsa;
      struct _gpgsv	gsv;
      struct _gprmc	rmc;
      struct _gpvtg	vtg;
      struct _gpzda	zda;
    } nmea;
  } gps;
} GPSData;


/*****************************************************************************
 *
 *	Global functions declaration
 *
 *****************************************************************************/
extern int	pacy_getinfo(struct _pacy *, FILE *);
extern void	pacy_raw2std(struct _pacy *, GPSInfo *);
extern int	nmea_getinfo(struct _nmea *, FILE *);
extern int	nmea_raw2std(struct _nmea *, GPSInfo *);

#endif	/* _GPS_SUB_H */
