/*****************************************************************************
 *
 * FILE:	griffin.h
 * DESCRIPTION:	GPS (Global Positioning System) library header file
 * DATE:	Thu, Dec 20 2001
 * UPDATED:	Sun, Sep 21 2003
 * AUTHOR:	Kouichi ABE (WALL)
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2003 Kouichi ABE (WALL), All rights reserved.
 * NOTE:	Supported GPS receiver
 *		1. "SONY PACY-CNV10 (PC-Card)" GPS receiver
 *		2. "DATATEC SPEED SENSOR (GPex)" GPS receiver
 *		3. "NMEA-0183 format" GPS receiver
 *
 *		   GP/ex SPECIFICATION:
 *			communication	asynchronized full-duplex
 *			speed		4800bps
 *			data		8bit
 *			stop bit	1bit
 *			parity		none
 *			output		ASCII code
 * LICENSE:
 *
 *  Copyright (C) 2001-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: griffin.h,v 1.3 2003/09/21 11:16:50 kouichi Exp $
 *
 *****************************************************************************/

#ifndef	_GRIFFIN_H
#define	_GRIFFIN_H

#include <stdio.h>
#if	TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif	/* TM_IN_SYS_TIME */


/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/

/*
 * Numeric release version identifier:
 * MNNFFPPS: major minor fix patch status
 * The status nibble has one of the values 0 for development,
 * 1 to e for betas 1 to 14, and f for release.
 * The patch level is exactly that.
 */
#define	GGL_VERSION_NUMBER	0x10002010L
#define	GGL_VERSION_TEXT	"Griffin 1.0.2 (2003/09/21)"
#define	GGL_VERSION_TEXT_SHORT	"Griffin/1.0.2"
#define	GGL_VERSION_TEXT_LONG	"Griffin GPS Library 1.0.2, Sun, Sep 21 2003"

/* Data Format Type */
typedef enum {
  GDF_UNKNOWN,
  GDF_SONY_PACY,	/* SONY PACY-CNV10 */
  GDF_NMEA0183		/* NMEA-0183 format */
} GPSDataFormat;

typedef struct {
  struct tm	timestamp;	/* ¬	*/
  char		ns;		/* ̰ޡNˡޡS	*/
  struct {
    long long	value;		/* ١ñ̡	*/
    int		degrees;	/* ١١		*/
    int		minutes;	/* ١ʬ: 1/60 deg	*/
    int		seconds;	/* ١á: 1/60 min	*/
    int		fraction;	/* ١ü		*/
  } latitude;
  char		ew;		/* СWˡСE	*/
  struct {
    long long	value;		/* ١ñ̡	*/
    int		degrees;	/* ١١		*/
    int		minutes;	/* ١ʬ: 1/60 deg	*/
    int		seconds;	/* ١á: 1/60 min	*/
    int		fraction;	/* ١ü		*/
  } longitude;
  double	velocity;	/* ®			*/
  double	direction;	/* 			*/
  double	altitude;	/* ɸ			*/
  /**/
  double	dop;		/* ١dilution of precision*/
  char		status;		/* ¬̾ 0:̤¬ 1:¬	*/
				/* 2: ǥե󥷥¬	*/
  int		satellite;	/* ѱ		*/
} GPSInfo;


/*****************************************************************************
 *
 *	Global functions declaration
 *
 *****************************************************************************/
extern int	gpsInit(GPSDataFormat, const char *);
extern void	gpsQuit(void);
extern int	gpsGetInfo(GPSInfo *);
extern void	gpsPrintInfo(GPSInfo *, FILE *);

#endif	/* _GRIFFIN_H */
