/*****************************************************************************
 *
 * FILE:	nmea0183.c
 * DESCRIPTION:	GPS functions for NMEA-0183 Format
 * DATE:	Wed, Jan 30 2002
 * UPDATED:	Sat, Sep 20 2003
 * AUTHOR:	Kouichi ABE (WALL)
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002-2003 Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (C) 2002-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: nmea0183.c,v 1.6 2003/09/21 11:16:50 kouichi Exp $
 *
 *****************************************************************************/

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <limits.h>
#include <math.h>
#include "griffin.h"
#include "gps-sub.h"


/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int	get_gpgga_info(struct _gpgga *, char *);
static int	get_gprmc_info(struct _gprmc *, char *);
static int	get_gpgll_info(struct _gpgll *, char *);
static int	get_gpgsa_info(struct _gpgsa *, char *);
static int	get_gpgsv_info(struct _gpgsv *, char *);
static int	get_gpvtg_info(struct _gpvtg *, char *);
static int	get_gpzda_info(struct _gpzda *, char *);
static int	mk_timestamp(struct _nmea *, GPSInfo *);
static double	coordinates_degree(double);
static double	coordinates_minutes(double);
static double	coordinates_seconds(double);
static int	mk_position(struct _nmea *, GPSInfo *);
#if	NMEA_DEBUG
static void	nmea_debug(struct _nmea *);
#endif	/* NMEA_DEBUG */


/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	nmea_getinfo
 * Description:		get GPS information with NMEA format
 * Arguments:		NMEA structure
 * Return Value:	Successful: 0, otherwise: -1
 */
#if	1
int
nmea_getinfo(nmea, fin)
	register struct _nmea *	nmea;
	register FILE *		fin;
{
  static char	saved[GBUFSIZE];
  char		buf[GBUFSIZE];
  enum	{ S_READY = -1, S_READING, S_FINISH } status;

  /* fixups */
  memset(buf, 0, sizeof(buf));
  if (saved[0] == '$') {
    memcpy(buf, saved, sizeof(buf));
  }
  else {
    fgets(buf, sizeof(buf), fin);
  }

  /* data reading */
  status = S_READY;
  do {
    memcpy(saved, buf, sizeof(saved));

    if (status == S_READY && strncmp(buf, "$GPGGA", 6) != 0) {
      continue;
    }

    if (strncmp(buf, "$GPGGA", 6) == 0) {
      memset((char *)&nmea->gga, 0, sizeof(nmea->gga));
      strncpy(nmea->gga.rawdata, buf, sizeof(nmea->gga.rawdata) - 1);
      if (get_gpgga_info(&nmea->gga, buf) == -1) {
	continue;
      }
      switch (status) {
	case S_READY:
	  status = S_READING;
	  break;
	case S_READING:
	  status = S_FINISH;
	  break;
	case S_FINISH:
	default:
	  status = S_READY;
	  break;
      }
    }
    else if (strncmp(buf, "$GPGLL", 6) == 0) {
      memset((char *)&nmea->gll, 0, sizeof(nmea->gll));
      strncpy(nmea->gll.rawdata, buf, sizeof(nmea->gll.rawdata) - 1);
      get_gpgll_info(&nmea->gll, buf);
    }
    else if (strncmp(buf, "$GPGSA", 6) == 0) {
      memset((char *)&nmea->gsa, 0, sizeof(nmea->gsa));
      strncpy(nmea->gsa.rawdata, buf, sizeof(nmea->gsa.rawdata) - 1);
      get_gpgsa_info(&nmea->gsa, buf);
    }
    else if (strncmp(buf, "$GPGSV", 6) == 0) {
      memset((char *)&nmea->gsv, 0, sizeof(nmea->gsv));
      strncpy(nmea->gsv.rawdata, buf, sizeof(nmea->gsv.rawdata) - 1);
      get_gpgsv_info(&nmea->gsv, buf);
    }
    else if (strncmp(buf, "$GPRMC", 6) == 0) {
      memset((char *)&nmea->rmc, 0, sizeof(nmea->rmc));
      strncpy(nmea->rmc.rawdata, buf, sizeof(nmea->rmc.rawdata) - 1);
      get_gprmc_info(&nmea->rmc, buf);
    }
    else if (strncmp(buf, "$GPVTG", 6) == 0) {
      memset((char *)&nmea->vtg, 0, sizeof(nmea->vtg));
      strncpy(nmea->vtg.rawdata, buf, sizeof(nmea->vtg.rawdata) - 1);
      get_gpvtg_info(&nmea->vtg, buf);
    }
    else if (strncmp(buf, "$GPZDA", 6) == 0) {
      memset((char *)&nmea->zda, 0, sizeof(nmea->zda));
      strncpy(nmea->zda.rawdata, buf, sizeof(nmea->zda.rawdata) - 1);
      get_gpzda_info(&nmea->zda, buf);
    }
    else {
    }
  } while (status != S_FINISH && fgets(buf, sizeof(buf), fin) != NULL);

  return (status == S_FINISH ? 0 : status);
}
#else
int
nmea_getinfo(nmea, fin)
	register struct _nmea *	nmea;
	register FILE *		fin;
{
  char	buf[GBUFSIZE];
  enum	{ S_READY = -1, S_READING, S_FINISH } status;

  for (status = S_READY;
       status != S_FINISH && fgets(buf, sizeof(buf), fin) != NULL; ) {
    if (status == S_READY && strncmp(buf, "$GPGGA", 6) != 0) {
      continue;
    }

    if (strncmp(buf, "$GPGGA", 6) == 0) {
      memset((char *)&nmea->gga, 0, sizeof(nmea->gga));
      strncpy(nmea->gga.rawdata, buf, sizeof(nmea->gga.rawdata) - 1);
      if (get_gpgga_info(&nmea->gga, buf) == -1) {
	continue;
      }
      switch (status) {
	case S_READY:
	  status = S_READING;
	  break;
	case S_READING:
	  status = S_FINISH;
	  break;
	case S_FINISH:
	default:
	  status = S_READY;
	  break;
      }
    }
    else if (strncmp(buf, "$GPGLL", 6) == 0) {
      memset((char *)&nmea->gll, 0, sizeof(nmea->gll));
      strncpy(nmea->gll.rawdata, buf, sizeof(nmea->gll.rawdata) - 1);
      get_gpgll_info(&nmea->gll, buf);
    }
    else if (strncmp(buf, "$GPGSA", 6) == 0) {
      memset((char *)&nmea->gsa, 0, sizeof(nmea->gsa));
      strncpy(nmea->gsa.rawdata, buf, sizeof(nmea->gsa.rawdata) - 1);
      get_gpgsa_info(&nmea->gsa, buf);
    }
    else if (strncmp(buf, "$GPGSV", 6) == 0) {
      memset((char *)&nmea->gsv, 0, sizeof(nmea->gsv));
      strncpy(nmea->gsv.rawdata, buf, sizeof(nmea->gsv.rawdata) - 1);
      get_gpgsv_info(&nmea->gsv, buf);
    }
    else if (strncmp(buf, "$GPRMC", 6) == 0) {
      memset((char *)&nmea->rmc, 0, sizeof(nmea->rmc));
      strncpy(nmea->rmc.rawdata, buf, sizeof(nmea->rmc.rawdata) - 1);
      get_gprmc_info(&nmea->rmc, buf);
    }
    else if (strncmp(buf, "$GPVTG", 6) == 0) {
      memset((char *)&nmea->vtg, 0, sizeof(nmea->vtg));
      strncpy(nmea->vtg.rawdata, buf, sizeof(nmea->vtg.rawdata) - 1);
      get_gpvtg_info(&nmea->vtg, buf);
    }
    else if (strncmp(buf, "$GPZDA", 6) == 0) {
      memset((char *)&nmea->zda, 0, sizeof(nmea->zda));
      strncpy(nmea->zda.rawdata, buf, sizeof(nmea->zda.rawdata) - 1);
      get_gpzda_info(&nmea->zda, buf);
    }
    else {
    }
  }

#if	NMEA_DEBUG
  nmea_debug(nmea);
#endif	/* NMEA_DEBUG */

  return (status == S_FINISH ? 0 : status);
}
#endif

/*
 * Function Name:	get_gpgga_info
 * Description:		get GPGGA formatted GPS data
 * Arguments:		GGA structure, data
 * Return Value:	none
 */
#define	GGA_ARGS		(15)
#define	GGA_TIME		(1)
#define	GGA_LATITUDE		(2)
#define	GGA_NS			(3)
#define	GGA_LONGITUDE		(4)
#define	GGA_EW			(5)
#define	GGA_QUALITY		(6)
#define	GGA_NUM			(7)
#define	GGA_DOP			(8)
#define	GGA_ALTITUDE		(9)
#define	GGA_ALTITUDE_UNIT	(10)
#define	GGA_GEOID		(11)
#define	GGA_GEOID_UNIT		(12)
#define	GGA_DGPS_DATA_AGE	(13)
#define	GGA_DGPS_STATION_ID	(14)
static int
get_gpgga_info(gga, data)
	register struct _gpgga *	gga;
	char *				data;
{
  char *	args[GGA_ARGS];
  char **	ap;
  char *	p;
  int		n;

  for (n = 0, ap = args; (*ap = strsep(&data, ",")) != NULL; ) {
    n++;
    if (++ap >= &args[GGA_ARGS]) {
      break;
    }
  }
  if (n < GGA_ARGS - 1) {
    return -1;
  }

  strncpy(gga->time, args[GGA_TIME], sizeof(gga->time) - 1);
  strncpy(gga->latitude, args[GGA_LATITUDE], sizeof(gga->latitude) - 1);
  gga->ns		= args[GGA_NS][0];
  strncpy(gga->longitude, args[GGA_LONGITUDE], sizeof(gga->longitude) - 1);
  gga->ew		= args[GGA_EW][0];
  gga->quality		= args[GGA_QUALITY][0];
  gga->num		= atoi(args[GGA_NUM]);
  gga->dop		= atof(args[GGA_DOP]);
  gga->altitude		= atof(args[GGA_ALTITUDE]);
  gga->altitude_unit	= args[GGA_ALTITUDE_UNIT][0];
  gga->geoid		= atof(args[GGA_GEOID]);
  gga->geoid_unit	= args[GGA_GEOID_UNIT][0];
  if (n == GGA_ARGS) {
    gga->dgps_data_age	= atof(args[GGA_DGPS_DATA_AGE]);
    p = strchr(args[GGA_DGPS_STATION_ID], '*');
    if (p != NULL) {
      *p++ = EOL;
      strncpy(gga->dgps_station_id, args[GGA_DGPS_STATION_ID],
  	      sizeof(gga->dgps_station_id) - 1);
      gga->checksum[0] = '*';
      strncpy(&gga->checksum[1], p, 2);
    }
  }
  else {
    p = strchr(args[GGA_DGPS_DATA_AGE], '*');
    if (p != NULL) {
      *p++ = EOL;
      gga->dgps_data_age = atof(args[GGA_DGPS_DATA_AGE]);
      gga->checksum[0] = '*';
      strncpy(&gga->checksum[1], p, 2);
    }
  }

  return 0;
}

/*
 * Function Name:	get_gpgll_info
 * Description:		get GPGLL formatted GPS data
 * Arguments:		GLL structure, data
 * Return Value:	none
 */
#define	GLL_ARGS		(8)
#define	GLL_LATITUDE		(1)
#define	GLL_NS			(2)
#define	GLL_LONGITUDE		(3)
#define	GLL_EW			(4)
#define	GLL_TIME		(5)
#define	GLL_STATUS		(6)
#define	GLL_MODE		(7)
static int
get_gpgll_info(gll, data)
	register struct _gpgll *	gll;
	char *				data;
{
  char *	args[GLL_ARGS];
  char **	ap;
  char *	p;
  int		n;

  for (n = 0, ap = args; (*ap = strsep(&data, ",")) != NULL; ) {
    n++;
    if (++ap >= &args[GLL_ARGS]) {
      break;
    }
  }
  if (n < GLL_ARGS - 1) {
    return -1;
  }

  strncpy(gll->latitude, args[GLL_LATITUDE], sizeof(gll->latitude) - 1);
  gll->ns	= args[GLL_NS][0];
  strncpy(gll->longitude, args[GLL_LONGITUDE], sizeof(gll->longitude) - 1);
  gll->ew	= args[GLL_EW][0];
  strncpy(gll->time, args[GLL_TIME], sizeof(gll->time) - 1);
  if (n == GLL_ARGS) {
    gll->status	= args[GLL_STATUS][0];
    gll->mode	= args[GLL_MODE][0];
    p = strchr(args[GLL_MODE], '*');
    if (p != NULL) {
      *p++ = EOL;
      gll->checksum[0] = '*';
      strncpy(&gll->checksum[1], p, 2);
    }
  }
  else {
    gll->status	= args[GLL_STATUS][0];
    p = strchr(args[GLL_STATUS], '*');
    if (p != NULL) {
      *p++ = EOL;
      gll->checksum[0] = '*';
      strncpy(&gll->checksum[1], p, 2);
    }
  }

  return 0;
}

/*
 * Function Name:	get_gpgsa_info
 * Description:		get GPGSA formatted GPS data
 * Arguments:		GSA structure, data
 * Return Value:	none
 */
#define	GSA_ARGS		(18)
#define	GSA_MEASURE_MODE	(1)
#define	GSA_MODE		(2)
#define	GSA_SATELLITE_NO	(3)
#define	GSA_PDOP		(15)
#define	GSA_HDOP		(16)
#define	GSA_VDOP		(17)
static int
get_gpgsa_info(gsa, data)
	register struct _gpgsa *	gsa;
	char *				data;
{
  char *	args[GSA_ARGS];
  char **	ap;
  int		n;
  int		i;
  char *	p;

  for (n = 0, ap = args; (*ap = strsep(&data, ",")) != NULL; ) {
    n++;
    if (++ap >= &args[GSA_ARGS]) {
      break;
    }
  }
  if (n < GSA_ARGS) {
    return -1;
  }

  gsa->measure_mode	= args[GSA_MEASURE_MODE][0];
  gsa->mode		= args[GSA_MODE][0];
  for (i = 0; i < MAX_GSA_SATELLITES; i++) {
    gsa->satellite_no[i] = atoi(args[GSA_SATELLITE_NO + i]);
  }
  gsa->pdop		= atof(args[GSA_PDOP]);
  gsa->hdop		= atof(args[GSA_HDOP]);
  p = strchr(args[GSA_VDOP], '*');
  if (p != NULL) {
    *p++ = EOL;
    gsa->vdop	     = atof(args[GSA_VDOP]);
    gsa->checksum[0] = '*';
    strncpy(&gsa->checksum[1], p, 2);
  }

  return 0;
}

/*
 * Function Name:	get_gpgsv_info
 * Description:		get GPGSV formatted GPS data
 * Arguments:		GSV structure, data
 * Return Value:	none
 */
#define	GSV_ARGS		(20)
#define	GSV_MESGS		(1)
#define	GSV_MSG_NO		(2)
#define	GSV_SATELLITE_NUM	(3)
#define	GSV_SATELLITES		(4)

static int
get_gpgsv_info(gsv, data)
	register struct _gpgsv *	gsv;
	char *				data;
{
  char *	args[GSV_ARGS];
  char **	ap;
  int		n;
  int		i;

  for (n = 0, ap = args; (*ap = strsep(&data, ",")) != NULL; ) {
    n++;
    if (++ap >= &args[GSV_ARGS]) {
      break;
    }
  }
  if (n < GSV_ARGS) {
    return -1;
  }

  gsv->mesgs		= atoi(args[GSV_MESGS]);
  gsv->msg_no		= atoi(args[GSV_MSG_NO]);
  gsv->satellite_num	= atoi(args[GSV_SATELLITE_NUM]);
  for (i = 0; i < (n / MAX_GSV_SATELLITES) - 1; i++) {
    int	base = GSV_SATELLITES * (i + 1);
    gsv->satellites[i].no	 = atoi(args[base]);
    gsv->satellites[i].angle	 = atoi(args[base + 1]);
    gsv->satellites[i].direction = atoi(args[base + 2]);
    if (i == MAX_GSV_SATELLITES - 1) {
      register char *	p;

      p = strchr(args[base + 3], '*');
      if (p != NULL) {
	*p++ = EOL;
	gsv->satellites[i].snr	= atoi(args[base + 3]);
	gsv->checksum[0]	= '*';
	strncpy(&gsv->checksum[1], p, 2);
      }
    }
    else {
      gsv->satellites[i].snr	 = atoi(args[base + 3]);
    }
  }

  return 0;
}

/*
 * Function Name:	get_gprmc_info
 * Description:		get GPRMC formatted GPS data
 * Arguments:		RMC structure, data
 * Return Value:	none
 */
#define	RMC_ARGS		(13)
#define	RMC_TIME		(1)
#define	RMC_STATUS		(2)
#define	RMC_LATITUDE		(3)
#define	RMC_NS			(4)
#define	RMC_LONGITUDE		(5)
#define	RMC_EW			(6)
#define	RMC_VELOCITY		(7)
#define	RMC_DIRECTION		(8)
#define	RMC_DATE		(9)
#define	RMC_MAGNETIC_DEVIATION	(10)
#define	RMC_MD_REVISE		(11)
#define	RMC_MODE		(12)
static int
get_gprmc_info(rmc, data)
	register struct _gprmc *	rmc;
	char *				data;
{
  char *	args[RMC_ARGS];
  char **	ap;
  int		n;

  for (n = 0, ap = args; (*ap = strsep(&data, ",")) != NULL; ) {
    n++;
    if (++ap >= &args[RMC_ARGS]) {
      break;
    }
  }
  if (n < RMC_ARGS - 1) {
    return -1;
  }

  strncpy(rmc->time, args[RMC_TIME], sizeof(rmc->time) - 1);
  rmc->status		= args[RMC_STATUS][0];
  strncpy(rmc->latitude, args[RMC_LATITUDE], sizeof(rmc->latitude) - 1);
  rmc->ns		= args[RMC_NS][0];
  strncpy(rmc->longitude, args[RMC_LONGITUDE], sizeof(rmc->longitude) - 1);
  rmc->ew		= args[RMC_EW][0];
  rmc->velocity		= atof(args[RMC_VELOCITY]);
  rmc->direction	= atof(args[RMC_DIRECTION]);
  strncpy(rmc->date, args[RMC_DATE], sizeof(rmc->date) - 1);
  strncpy(rmc->magnetic_deviation, args[RMC_MAGNETIC_DEVIATION],
	  sizeof(rmc->magnetic_deviation) - 1);
  if (n == RMC_ARGS) {
    rmc->md_revise	= args[RMC_MD_REVISE][0];
    rmc->mode		= args[RMC_MODE][0];
    strncpy(rmc->checksum, &args[RMC_MODE][1], sizeof(rmc->checksum) - 1);
  }
  else {
    rmc->md_revise = args[RMC_MD_REVISE][0];
    strncpy(rmc->checksum, &args[RMC_MD_REVISE][1], sizeof(rmc->checksum) - 1);
  }

  return 0;
}

/*
 * Function Name:	get_gpvtg_info
 * Description:		get GPVTG formatted GPS data
 * Arguments:		VTG structure, data
 * Return Value:	none
 */
#define	VTG_ARGS		(10)
#define	VTG_DIRECTION		(1)
#define	VTG_DIRECTION_UNIT	(2)
#define	VTG_MAGNETIC_DIRECTION	(3)
#define	VTG_MAGNETIC_DIR_UNIT	(4)
#define	VTG_VELOCITY1		(5)
#define	VTG_VELOCITY1_UNIT	(6)
#define	VTG_VELOCITY2		(7)
#define	VTG_VELOCITY2_UNIT	(8)
#define	VTG_MODE		(9)
static int
get_gpvtg_info(vtg, data)
	register struct _gpvtg *	vtg;
	char *				data;
{
  char *	args[VTG_ARGS];
  char **	ap;
  int		n;

  for (n = 0, ap = args; (*ap = strsep(&data, ",")) != NULL; ) {
    n++;
    if (++ap >= &args[VTG_ARGS]) {
      break;
    }
  }
  if (n < VTG_ARGS - 1) {
    return -1;
  }

  vtg->direction		= atof(args[VTG_DIRECTION]); 
  vtg->direction_unit		= args[VTG_DIRECTION_UNIT][0];
  vtg->magnetic_direction	= atof(args[VTG_MAGNETIC_DIRECTION]); 
  vtg->magnetic_dir_unit	= args[VTG_MAGNETIC_DIR_UNIT][0];
  vtg->velocity1		= atof(args[VTG_VELOCITY1]); 
  vtg->velocity1_unit		= args[VTG_VELOCITY1_UNIT][0];
  vtg->velocity2		= atof(args[VTG_VELOCITY2]); 
  vtg->velocity2_unit		= args[VTG_VELOCITY2_UNIT][0];
  if (n == VTG_ARGS) {
    vtg->mode			= args[VTG_MODE][0];
    strncpy(vtg->checksum, &args[VTG_MODE][1], sizeof(vtg->checksum) - 1);
  }

  return 0;
}

/*
 * Function Name:	get_gpzda_info
 * Description:		get GPZDA formatted GPS data
 * Arguments:		ZDA structure, data
 * Return Value:	none
 */
#define	ZDA_ARGS	(7)
#define	ZDA_TIME	(1)
#define	ZDA_DAY		(2)
#define	ZDA_MONTH	(3)
#define	ZDA_YEAR	(4)
#define	ZDA_HOUR	(5)
#define	ZDA_MIN		(6)
static int
get_gpzda_info(zda, data)
	register struct _gpzda *	zda;
	char *				data;
{
  char *	args[ZDA_ARGS];
  char **	ap;
  int		n;
  char *	p;

  for (n = 0, ap = args; (*ap = strsep(&data, ",")) != NULL; ) {
    n++;
    if (++ap >= &args[ZDA_ARGS]) {
      break;
    }
  }
  if (n < ZDA_ARGS) {
    return -1;
  }

  strncpy(zda->time,  args[ZDA_TIME],  sizeof(zda->time) - 1);
  strncpy(zda->day,   args[ZDA_DAY],   sizeof(zda->day) - 1);
  strncpy(zda->month, args[ZDA_MONTH], sizeof(zda->month) - 1);
  strncpy(zda->year,  args[ZDA_YEAR],  sizeof(zda->year) - 1);
  strncpy(zda->hour,  args[ZDA_HOUR],  sizeof(zda->hour) - 1);
  p = strchr(args[ZDA_MIN], '*');
  if (p != NULL) {
    *p++ = EOL;
    strncpy(zda->min, args[ZDA_MIN], sizeof(zda->min) - 1);
    zda->checksum[0] = '*';
    strncpy(&zda->checksum[1], p, 2);
  }

  return 0;
}

/*****************************************************************************/

/*
 * Function Name:	nmea_raw2std
 * Description:		convert GPS raw data to standard information data
 * Arguments:		GPS {raw,standard} information structure
 * Return Value:	none
 */
int
nmea_raw2std(nmea, gpsi)
	register struct _nmea *	nmea;
	register GPSInfo *	gpsi;
{
  if (mk_timestamp(nmea, gpsi) == -1 || mk_position(nmea, gpsi) == -1) {
    return -1;
  }

  /* ® */
  gpsi->velocity	= nmea->vtg.velocity2;
  /*  */
  gpsi->direction	= nmea->vtg.direction;
  /* ɸ */
  gpsi->altitude	= nmea->gga.altitude;

  /* ¬ */
  gpsi->dop		= nmea->gga.dop;
  /* ¬̾ */
  gpsi->status		= nmea->gga.quality;
  /* ѱ */
  if (nmea->gsv.satellite_num > 0) {
    gpsi->satellite = nmea->gsv.satellite_num;
  }
  else if (nmea->gga.num > 0) {
    gpsi->satellite = nmea->gga.num;
  }
  else {
    gpsi->satellite = 0;
  }

  return 0;
}

static int
mk_timestamp(nmea, gpsi)
	register struct _nmea *	nmea;
	register GPSInfo *	gpsi;
{
#define	WDAY(y,m,d)	(((y)+(y)/4-(y)/100+(y)/400+(13*(m)+8)/5+d)%7)
  static int	yday[] = {0, 31, 59, 90, 120, 151, 181, 212, 242, 273,
			  303, 334};
  int		hour, min, sec;
  int		year, month, day, leap;

  /* time */
  if (nmea->zda.time[0] != EOL) {
    sscanf(nmea->zda.time, "%2d%2d%2d", &hour, &min, &sec);
  }
  else if (nmea->gga.time[0] != EOL) {
    sscanf(nmea->gga.time, "%2d%2d%2d", &hour, &min, &sec);
  }
  else if (nmea->gll.time[0] != EOL) {
    sscanf(nmea->gll.time, "%2d%2d%2d", &hour, &min, &sec);
  }
  else if (nmea->rmc.time[0] != EOL) {
    sscanf(nmea->rmc.time, "%2d%2d%2d", &hour, &min, &sec);
  }
  else {
    return -1;
  }

  /* date */
  if (nmea->zda.year[0] != EOL && nmea->zda.month[0] != EOL &&
      nmea->zda.day[0] != EOL) {
    year  = atoi(nmea->zda.year);
    month = atoi(nmea->zda.month);
    day	  = atoi(nmea->zda.day);
  }
  else if (nmea->rmc.date[0] != EOL) {
    sscanf(nmea->rmc.date, "%2d%2d%2d", &day, &month, &year);
    year = year + (year > 72 ? 1900 : 2000);
  }
  else {
    return -1;
  }

  /* Time and Date */
  gpsi->timestamp.tm_year = year - 1900;
  gpsi->timestamp.tm_mon  = month - 1;
  gpsi->timestamp.tm_mday = day;
  leap = ((year % 4 == 0) && (year % 100 != 0)) || (year % 400 == 0) ? 1 : 0;
  if (month > 1) {
    gpsi->timestamp.tm_yday = yday[month] + day + leap;
  }
  else {
    gpsi->timestamp.tm_yday = yday[month] + day;
  }
  gpsi->timestamp.tm_wday	= WDAY(year,month,day);
  gpsi->timestamp.tm_hour	= hour;
  gpsi->timestamp.tm_min	= min;
  gpsi->timestamp.tm_sec	= sec;
  gpsi->timestamp.tm_isdst	= 0;
  gpsi->timestamp.tm_gmtoff	= DEF_OFFSET;
  gpsi->timestamp.tm_zone	= DEF_TIMEZONE;

  return 0;
}


static double
coordinates_degree(val)
	double	val;
{
  double	ret;
  int		deg  = (int)floor(val);
  int		min  = deg % 100;
  double	fmin = 0.0;	/* fracttional minitues */
  double	iptr = 0.0;

  fmin = modf(val, &iptr);
  deg -= min;
  deg /= 100;

  ret  = deg;
  ret += (double)((double)min / (double)60.0);
  ret += (double)((double)fmin / (double)60.0);

  return ret;
}

static double
coordinates_minutes(val)
	double	val;
{
  double	ret;
  int		deg  = (int)floor(val);
  int		min  = deg % 100;
  double	fmin = 0.0;	/* fracttional minitues */
  double	iptr = 0.0;

  fmin = modf(val, &iptr);

  ret  = (double)min;
  ret += fmin;

  return ret;
}

static double
coordinates_seconds(val)
	double	val;
{
  double	ret;
  double	fmin = 0.0;	/* fracttional minitues */
  double	iptr = 0.0;

  fmin = modf(val, &iptr);

  ret  = fmin * (double)60.0;

  return ret;
}

static int
mk_position(nmea, gpsi)
	register struct _nmea *	nmea;
	register GPSInfo *	gpsi;
{
  int		deg, min, sec, frac;
  double	f;
  double	latitude;
  double	longitude;
  char		buf[24];

  /* ̰ޡ	*/
  if (nmea->gll.ns != EOL) {
    gpsi->ns = nmea->gll.ns;
  }
  else if (nmea->gga.ns != EOL) {
    gpsi->ns = nmea->gga.ns;
  }
  else if (nmea->rmc.ns != EOL) {
    gpsi->ns = nmea->rmc.ns;
  }
  else {
    return -1;
  }

  /*  */
  if (nmea->gll.latitude[0] != EOL) {
    latitude = strtod(nmea->gll.latitude, (char **)NULL);
  }
  else if (nmea->gga.latitude[0] != EOL) {
    latitude = strtod(nmea->gga.latitude, (char **)NULL);
  }
  else if (nmea->rmc.latitude[0] != EOL) {
    latitude = strtod(nmea->rmc.latitude, (char **)NULL);
  }
  else {
    return -1;
  }

  deg  = (int)floor(coordinates_degree(latitude));
  min  = (int)floor(coordinates_minutes(latitude));
  f    = coordinates_seconds(latitude);
  sec  = (int)floor(f);
  frac = (int)((f - (double)sec) * 1000);

  memset(buf, 0, sizeof(buf));
  snprintf(buf, sizeof(buf), "%02d%02d%02d%04d", deg, min, sec, frac);
  gpsi->latitude.value		= strtoll(buf, (char **)NULL, 10);
  gpsi->latitude.degrees	= deg;		/* ١١	*/
  gpsi->latitude.minutes	= min;		/* ١ʬ	*/
  gpsi->latitude.seconds	= sec;		/* ١á	*/
  gpsi->latitude.fraction	= frac;		/* ١1/1000á*/

  /* С	*/
  if (nmea->gll.ew != EOL) {
    gpsi->ew = nmea->gll.ew;
  }
  else if (nmea->gga.ew != EOL) {
    gpsi->ew = nmea->gga.ew;
  }
  else if (nmea->rmc.ew != EOL) {
    gpsi->ew = nmea->rmc.ew;
  }
  else {
    return -1;
  }

  /*  */
  if (nmea->gll.longitude[0] != EOL) {
    longitude = strtod(nmea->gll.longitude, (char **)NULL);
  }
  else if (nmea->gga.longitude[0] != EOL) {
    longitude = strtod(nmea->gga.longitude, (char **)NULL);
  }
  else if (nmea->rmc.longitude[0] != EOL) {
    longitude = strtod(nmea->rmc.longitude, (char **)NULL);
  }
  else {
    return -1;
  }

  deg  = (int)floor(coordinates_degree(longitude));
  min  = (int)floor(coordinates_minutes(longitude));
  f    = coordinates_seconds(longitude);
  sec  = (int)floor(f);
  frac = (int)((f - (double)sec) * 1000);

  memset(buf, 0, sizeof(buf));
  snprintf(buf, sizeof(buf), "%02d%02d%02d%04d", deg, min, sec, frac);
  gpsi->longitude.value		= strtoll(buf, (char **)NULL, 10);
  gpsi->longitude.degrees	= deg;		/* ١١	*/
  gpsi->longitude.minutes	= min;		/* ١ʬ	*/
  gpsi->longitude.seconds	= sec;		/* ١á	*/
  gpsi->longitude.fraction	= frac;		/* ١1/1000á*/

  return 0;
}


#if	NMEA_DEBUG
static void
nmea_debug(gpsi)
	struct _nmea *	gpsi;
{
  fprintf(stdout, "\n[GPGGA]\n");
  fprintf(stdout, ":\t%s\n", gpsi->gga.time);
  fprintf(stdout, ":\t%s(%c)\n", gpsi->gga.latitude, gpsi->gga.ns);
  fprintf(stdout, ":\t%s(%c)\n", gpsi->gga.longitude, gpsi->gga.ew);
  fprintf(stdout, ":\t%c\n", gpsi->gga.quality);
  fprintf(stdout, ":\t%d\n", gpsi->gga.num);
  fprintf(stdout, " DOP:\t%.3f\n", gpsi->gga.dop);
  fprintf(stdout, "ɸ:\t%.1f(%c)\n",
			gpsi->gga.altitude, gpsi->gga.altitude_unit);
  fprintf(stdout, "geoid:\t%.1f(%c)\n",
			gpsi->gga.geoid, gpsi->gga.geoid_unit);


  fprintf(stdout, "\n[GPVTG]\n");
  fprintf(stdout, ":\t%.1f(%c)\n", gpsi->vtg.direction, gpsi->vtg.dflag);
  fprintf(stdout, ":\t%.1f(%c)\n",
		gpsi->vtg.magnetic_direction, gpsi->vtg.mflag);
  fprintf(stdout, "®:\t%.1f(%c)\n", gpsi->vtg.velocity1, gpsi->vtg.v1flag);
  fprintf(stdout, "®:\t%.1f(%c)\n", gpsi->vtg.velocity2, gpsi->vtg.v2flag);

  fprintf(stdout, "\n[GPRMC]\n");
  fprintf(stdout, ":\t%s\n", gpsi->rmc.time);
  fprintf(stdout, ":\t%c\n", gpsi->rmc.status);
  fprintf(stdout, ":\t%s(%c)\n", gpsi->rmc.latitude, gpsi->rmc.ns);
  fprintf(stdout, ":\t%s(%c)\n", gpsi->rmc.longitude, gpsi->rmc.ew);
  fprintf(stdout, "®:\t%.1f\n", gpsi->rmc.velocity);
  fprintf(stdout, ":\t%.1f\n", gpsi->rmc.direction);
  fprintf(stdout, ":\t%s\n", gpsi->rmc.date);
  fprintf(stdout, "к:\t%s(%c)\n",
		gpsi->rmc.magnetic_deviation, gpsi->rmc.md_revise);
  fprintf(stdout, "\t[%c](%s)\n", gpsi->rmc.mode, gpsi->rmc.checksum);

  fprintf(stdout, "\n[GPZDA]\n");
  fprintf(stdout, ":\t%s\n", gpsi->zda.time);
  fprintf(stdout, ":\t%s\n", gpsi->zda.day);
  fprintf(stdout, ":\t%s\n", gpsi->zda.month);
  fprintf(stdout, "ǯ:\t%s\n", gpsi->zda.year);
  fprintf(stdout, ":\t%s\n", gpsi->zda.hour);
  fprintf(stdout, "ʬ:\t%s\n\n", gpsi->zda.min);
}
#endif	/* NMEA_DEBUG */

